pub usingnamespace @import("ingble.zig");

pub usingnamespace @import("gatt_client_async.zig");

const print = platform_printf;

extern fn print_uuid(uuid: [*c]const u8) void;
extern fn print_properties(v: u16) void;

var services: [10]gatt_client_service_t  = undefined;
var service_num: usize = undefined;

var characteristics: [10]gatt_client_characteristic_t  = undefined;
var characteristic_num: usize = undefined;

fn print_service(service: *gatt_client_service_t) void {
    print("\nSERVICE: ");
    print_uuid(&service.uuid128);
    print("\nHANDLE RANGE: %d - %d\n\n", service.start_group_handle, service.end_group_handle);
}

fn print_characteristic(characteristic: *gatt_client_characteristic_t) void {
    print("\n    CHARACTERISTIC: ");
    print_uuid(&characteristic.uuid128);
    print("\n        HANDLE RANGE: %d - %d\n", characteristic.start_handle, characteristic.end_handle);
    print(  "        VALUE HANDLE: %d\n", characteristic.value_handle);
    print(  "        PROPERTIES  : "); print_properties(characteristic.properties);
}

fn print_descriptor(descriptor: [*c] const gatt_client_characteristic_descriptor_t) void {
    print("\n        DESCRIPTOR: ");
    print_uuid(&descriptor.*.uuid128);
    print("\n        HANDLE: %d\n\n", descriptor.*.handle);
}

fn zig_aysnc_discover() void {
    service_num = 0;
    _ = all_services_discoverer.start(0);
    while (all_services_discoverer.next()) |service| {
        if (service_num < services.len) {
            services[service_num] = service;
            service_num += 1;
        }
    }

    print("========== GATT PROFILE DUMP ==========\n\n");

    var index: usize = 0;
    while (index < service_num) : ( index += 1) {
        print_service(&services[index]);

        characteristic_num = 0;
        _ = all_characteristics_discoverer.start(0, services[index].start_group_handle,
            services[index].end_group_handle);
        while (all_characteristics_discoverer.next()) |characteristic| {
            if (characteristic_num < characteristics.len) {
                characteristics[characteristic_num] = characteristic;
                characteristic_num += 1;
            }
        }

        var char_index: usize = 0;
        while (char_index < characteristic_num) : (char_index += 1) {
            print_characteristic(&characteristics[char_index]);

            if ((characteristics[char_index].properties & 0x2) != 0) {
                print("\n        VALUE       :");
                show_char_value(0, characteristics[char_index].value_handle);
            }
            print("\n");

            _ = all_descriptors_discoverer.start(0, &characteristics[char_index]);
            while (all_descriptors_discoverer.next()) |descriptor| {
                print_descriptor(&descriptor);
            }
        }
    }

    print("=====================================\n");
}

fn demo() void {
    var service = service_discoverer.discover_16(0,
        SIG_UUID_SERVICE_GENERIC_ACCESS) orelse return;
    var char_name = characteristics_discoverer.discover_16(0,
        service.start_group_handle, service.end_group_handle,
        SIG_UUID_CHARACT_GAP_DEVICE_NAME) orelse return;
    print("\nDEVICE_NAME       :");
    show_char_value(0, char_name.value_handle);
}

var disc_frame: @Frame(zig_aysnc_discover) = undefined;

var demo_frame: @Frame(demo) = undefined;

export fn zig_discover() void {
    disc_frame = async zig_aysnc_discover();
    //demo_frame = async demo();
}

extern fn print_hex_table(value: [*c] const u8, len: c_int) void;

fn show_char_value(conn_handle: hci_con_handle_t, value_handle: u16) void {
    var value_size: u16 = 0;
    if (value_of_characteristic_reader.read(conn_handle, value_handle, &value_size)) |value| {
        print("\n");
        print_hex_table(value, @as(c_int, value_size));
    }
}

var read_value_frame: @Frame(show_char_value) = undefined;
export fn read_char_value(conn_handle: hci_con_handle_t, value_handle: u16) void {
    read_value_frame = async show_char_value(conn_handle, value_handle);
}

fn a_write_char_value(conn_handle: hci_con_handle_t, value_handle: u16,
                 value_size: u16, value: [*c] u8) void {
    var r = await async value_of_characteristic_writer.write(conn_handle, value_handle, value_size, value);
    print("write_characteristic_value COMPLETE: %d\n", value_of_characteristic_writer.status);
    if (!r) {
        _ = gap_disconnect(conn_handle);
    }
}

var write_value_frame: @Frame(a_write_char_value) = undefined;
export fn write_char_value(conn_handle: hci_con_handle_t, value_handle: u16,
                 value_size: u16, value: [*c] u8) void {
    write_value_frame = async a_write_char_value(conn_handle, value_handle, value_size, value);
}

const std = @import("std");

pub fn panic(message: []const u8, stack_trace: ?*std.builtin.StackTrace) noreturn {
    print("\n!KERNEL PANIC!\n");
    for (message) |value| {
        print("%c", value);
    }
    while(true) {}
}